library form_control;

import 'package:flutter/material.dart';
import 'package:mobile_ui/src/controls/row_spacing.dart';
import 'package:mobile_ui/src/controls/ui_spacing.dart';

part 'checkbox_form_field.dart';
part 'switch_form_field.dart';
part 'radio_list_form_field.dart';

typedef FieldValueValidator<T> = String Function(
    T value, String error, String errorMsg);
enum ControlPosition { leading, trailing }
enum FieldOrientation { vertical, horizontal }

class FormFieldBase<T> extends FormField<T> {
  final bool isRequired;
  final String labelText;
  final String hintText;
  final EdgeInsetsGeometry contentPadding;
  final EdgeInsetsGeometry margin;
  final double spacingFlactor;
  final bool enabled;
  final bool readOnly;
  final TextAlign labelAlign;
  final FieldOrientation orientation;
  final ControlPosition controlPosition;
  final FieldEditingController<T> controller;

  FormFieldBase({
    Key key,
    @required FormFieldBuilder<T> builder,
    FieldValueValidator<T> validator,
    T initialValue,
    this.controller,
    this.isRequired,
    this.labelText,
    this.hintText,
    this.contentPadding = const EdgeInsets.all(8),
    this.margin = const EdgeInsets.symmetric(vertical: 8),
    this.spacingFlactor = 2,
    this.enabled,
    this.readOnly,
    this.labelAlign = TextAlign.left,
    this.orientation = FieldOrientation.vertical,
    //Only valid when the direction is equal to horizontal
    this.controlPosition = ControlPosition.trailing,
    AutovalidateMode autovalidateMode,
    FormFieldSetter<T> onSaved,
  })  : assert(initialValue == null || controller == null),
        super(
            builder: builder,
            initialValue: initialValue,
            onSaved: onSaved,
            enabled: enabled,
            autovalidateMode: autovalidateMode,
            validator: (value) {
              String errorCode;
              String errorMsg;
              if (value == null) {
                if (isRequired == true) {
                  errorCode = 'required';
                  errorMsg = '${labelText ?? "This field"} is required.';
                }
              }
              if (validator != null)
                return validator(value, errorCode, errorMsg);
              return errorMsg;
            });

  Widget buildField(FormFieldState<T> field, Widget child) {
    if (orientation == FieldOrientation.vertical) {
      return buildVField(field, child);
    } else {
      return buildHField(field, child);
    }
  }

  Widget buildHField(FormFieldState<T> field, Widget child) {
    if (labelText != null && labelText.isNotEmpty) {
      var children = <Widget>[Expanded(child: buildLabelText(field))];
      if (controlPosition == ControlPosition.leading) {
        children.insert(0, child);
      } else {
        children.add(child);
      }
      return Padding(
          padding: margin,
          child: Row(
            mainAxisAlignment: MainAxisAlignment.start,
            children: children,
          ));
    }
    return child;
  }

  Widget buildVField(FormFieldState<T> field, Widget child) {
    var children = <Widget>[];
    if (labelText != null && labelText.isNotEmpty) {
      children.addAll([
        buildLabelText(field),
        RowSpacing(spacingFlactor: spacingFlactor),
      ]);
    }
    return Padding(
        padding: margin,
        child: Column(
            crossAxisAlignment: CrossAxisAlignment.stretch,
            children: [...children, child]));
  }

  Widget buildLabelText(FormFieldState<T> field) {
    final theme = Theme.of(field.context);
    return Text(
      labelText ?? '',
      textAlign: labelAlign,
      style: enabled
          ? theme.textTheme.subtitle1
          : theme.textTheme.subtitle1.copyWith(color: theme.disabledColor),
    );
  }

  Widget buildIcon(
      FormFieldState<T> field, IconData icon, GestureTapCallback onTap) {
    final color = Theme.of(field.context).brightness == Brightness.light
        ? Colors.grey.shade700
        : Colors.white70;
    final iconWidget = Icon(
      icon,
      color: color,
    );
    if (onTap != null) {
      return InkWell(
        onTap: onTap,
        child: iconWidget,
      );
    }
    return iconWidget;
  }

  Widget buildDecorator(
      FormFieldState<T> field, Widget child, bool hasWhiteBackgroud) {
    final effectiveDecoration = InputDecoration(
      contentPadding: contentPadding,
    ).copyWith(
        hintText: hintText,
        border: OutlineInputBorder(borderSide: BorderSide.none));

    final inputDecorator = InputDecorator(
        decoration: effectiveDecoration.copyWith(errorText: field.errorText),
        isEmpty: field.value == null,
        child: child);
    if (hasWhiteBackgroud) {
      return Container(color: Colors.white, child: inputDecorator);
    }
    return inputDecorator;
  }

  @override
  FormFieldStateBase<T> createState() => FormFieldStateBase<T>();
}

class FormFieldStateBase<T> extends FormFieldState<T> {
  FieldEditingController _controller;

  FieldEditingController get _effectiveController =>
      widget.controller ?? _controller;

  @override
  FormFieldBase<T> get widget => super.widget;

  @override
  void initState() {
    super.initState();
    if (widget.controller == null) {
      _controller = FieldEditingController(widget.initialValue);
    } else {
      widget.controller.addListener(_handleControllerChanged);
    }
  }

  @override
  void didUpdateWidget(FormFieldBase<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      oldWidget.controller?.removeListener(_handleControllerChanged);
      widget.controller?.addListener(_handleControllerChanged);

      if (oldWidget.controller != null && widget.controller == null) {
        _controller = FieldEditingController(oldWidget.controller.value);
      }
      if (widget.controller != null) {
        setValue(widget.controller.value);
        if (oldWidget.controller == null) _controller = null;
      }
    }
  }

  @override
  void dispose() {
    widget.controller?.removeListener(_handleControllerChanged);
    super.dispose();
  }

  @override
  void reset() {
    super.reset();
    setState(() {
      _effectiveController.value = widget.initialValue;
    });
  }

  void _handleControllerChanged() {
    if (_effectiveController.value != value) {
      didChange(_effectiveController.value);
    }
  }

  void changeValue(T v) {
    _effectiveController.value = v;
  }
}

class FieldEditingController<T> extends ValueNotifier<T> {
  FieldEditingController(T value) : super(value);

  FieldEditingController.fromValue(T value) : super(value);
}
